/* Controlling the TeleToyland Sandbox 3 (table zize) with a Joystick Shield for Local Shows

   The Joystick shield sits on top of the Arduino, and uses pins 2-6 for the buttons, and analog 0 & 1 for the joystick
   The pen up/down servo is connected to pin 9
   The X & Y motors use Pololu JRK 21v3 motor coltrollers with 5-turn 5K potentiometers
   See http://www.instructables.com/id/Internet-Arduino-Controlled-T-Slot-XY-Table/ for details
   The TX port is wired to the JRK motor controllers
   So, a tether is used to connect the Arduio + Joystick shield to the motor controllers
     - pin 9 for the servo, the TX pin, and Vcc/Gnd
   A 6V battery was used to power both the Arduino and motor controllers
   
   by Carl Sutter
   June 2013
 */
 
#include <Servo.h> 


int x_last;  // the previous  X & Y position
int y_last;

int x_joy;   // the X & Y position from the joystick
int y_joy;

const int servo_up = 45;     // servo angle for the pen to be up
const int servo_down = 105;  // servo angle for the pen to be down
Servo Zservo;                // create servo object to control a servo 

// Store the Arduino pin associated with each input
const byte PIN_BUTTON_SELECT = 2; // Select button is triggered when joystick is pressed
const byte PIN_BUTTON_RIGHT = 3;
const byte PIN_BUTTON_UP = 4;
const byte PIN_BUTTON_DOWN = 5;
const byte PIN_BUTTON_LEFT = 6;
const byte PIN_ANALOG_X = 0;
const byte PIN_ANALOG_Y = 1;

// move the X axis to the specified position
// See the Pololu JRK docs - this is a high res Pololu type command since we have 2 boards
void XMove(int x) {
  word target = x;
  //Serial.print("X: ");
  //Serial.print(x);
  Serial.write(0xAA);
  Serial.write(0x0B);
  Serial.write(0x40 + (target & 0x1F));
  Serial.write((target >> 5) & 0x7F);
}  

// move the Y axis to the specified position
void YMove(int y) {
  word target = y;
  //Serial.print(", Y: ");
  //Serial.println(y);
  Serial.write(0xAA);
  Serial.write(0x0C);
  Serial.write(0x40 + (target & 0x1F));
  Serial.write((target >> 5) & 0x7F);
}


// draw a grid of lines to wipe out the sandbox
void wipeout() {
    const int step_size = 40;
    int i;
    int yPos = 0;
    
    enqueue(1001, 0); // up
    enqueue(0, 0);
    enqueue(1002, 0); // down
    
    for (i=0; i<=(1000 - 2*step_size); i += 2*step_size) {
      enqueue(1000, yPos);
      
      yPos = i + step_size;
      enqueue(1000, yPos);

      enqueue(0, yPos);

      yPos = i + 2*step_size;
      enqueue(0, yPos);
    }

    enqueue(1001, 0); // up
    ProcessQueue();
    x_last = 0;
    y_last = 4095;
} // wipeout


// draw a Snoopy figure
void snoopy() {
    const int x_offset = 150;
  
    enqueue(1001, 0); // up
    enqueue(0*18 + x_offset, 42*18);
    enqueue(1002, 0); // down
    enqueue(1*18 + x_offset, 31*18);
    enqueue(4*18 + x_offset, 17*18);
    enqueue(6*18 + x_offset, 27*18);
    enqueue(10*18 + x_offset, 30*18);
    enqueue(11*18 + x_offset, 31*18);
    enqueue(14*18 + x_offset, 29*18);
    enqueue(14*18 + x_offset, 28*18);
    enqueue(13*18 + x_offset, 26*18);
    enqueue(11*18 + x_offset, 24*18);
    enqueue(10*18 + x_offset, 21*18);
    enqueue(10*18 + x_offset, 18*18);
    enqueue(13*18 + x_offset, 12*18);
    enqueue(12*18 + x_offset, 11*18);
    enqueue(11*18 + x_offset, 12*18);
    enqueue(9*18 + x_offset, 11*18);
    enqueue(11*18 + x_offset, 7*18);
    enqueue(14*18 + x_offset, 5*18);
    enqueue(17*18 + x_offset, 4*18);
    enqueue(18*18 + x_offset, 5*18);
    enqueue(18*18 + x_offset, 6*18);
    enqueue(15*18 + x_offset, 9*18);
    enqueue(16*18 + x_offset, 10*18);
    enqueue(18*18 + x_offset, 10*18);
    enqueue(19*18 + x_offset, 7*18);
    enqueue(18*18 + x_offset, 6*18);
    enqueue(18*18 + x_offset, 5*18);
    enqueue(19*18 + x_offset, 4*18);
    enqueue(24*18 + x_offset, 6*18);
    enqueue(27*18 + x_offset, 10*18);
    enqueue(27*18 + x_offset, 11*18);
    enqueue(26*18 + x_offset, 12*18);
    enqueue(22*18 + x_offset, 10*18);
    enqueue(21*18 + x_offset, 12*18);
    enqueue(23*18 + x_offset, 14*18);
    enqueue(25*18 + x_offset, 18*18);
    enqueue(25*18 + x_offset, 20*18);
    enqueue(23*18 + x_offset, 23*18);
    enqueue(19*18 + x_offset, 26*18);
    enqueue(18*18 + x_offset, 28*18);
    enqueue(18*18 + x_offset, 29*18);
    enqueue(20*18 + x_offset, 32*18);
    enqueue(24*18 + x_offset, 33*18);
    enqueue(30*18 + x_offset, 33*18);
    enqueue(35*18 + x_offset, 35*18);
    enqueue(37*18 + x_offset, 37*18);
    enqueue(38*18 + x_offset, 39*18);
    enqueue(38*18 + x_offset, 42*18);
    enqueue(37*18 + x_offset, 44*18);
    enqueue(34*18 + x_offset, 47*18);
    enqueue(28*18 + x_offset, 49*18);
    enqueue(21*18 + x_offset, 48*18);
    enqueue(14*18 + x_offset, 51*18);
    enqueue(8*18 + x_offset, 51*18);
    enqueue(6*18 + x_offset, 50*18);
    enqueue(0*18 + x_offset, 42*18);

    enqueue(1001, 0); // up
    enqueue(37*18 + x_offset, 42*18);
    enqueue(1002, 0); // down
    enqueue(39*18 + x_offset, 42*18);
    enqueue(40*18 + x_offset, 41*18);
    enqueue(40*18 + x_offset, 40*18);
    enqueue(39*18 + x_offset, 39*18);
    enqueue(37*18 + x_offset, 39*18);

    enqueue(1001, 0); // up
    enqueue(18*18 + x_offset, 41*18);
    enqueue(1002, 0); // down
    enqueue(18*18 + x_offset, 42*18);
    enqueue(19*18 + x_offset, 43*18);

    enqueue(1001, 0); // up
    enqueue(11*18 + x_offset, 31*18);
    enqueue(1002, 0); // down
    enqueue(12*18 + x_offset, 34*18);
    enqueue(12*18 + x_offset, 39*18);
    enqueue(11*18 + x_offset, 42*18);

    enqueue(1001, 0); // up
    enqueue(4*18 + x_offset, 45*18);
    enqueue(1002, 0); // down
    enqueue(1*18 + x_offset, 41*18);
    enqueue(2*18 + x_offset, 31*18);
    enqueue(4*18 + x_offset, 28*18);
    enqueue(6*18 + x_offset, 28*18);
    enqueue(10*18 + x_offset, 31*18);
    enqueue(11*18 + x_offset, 34*18);
    enqueue(11*18 + x_offset, 39*18);
    enqueue(10*18 + x_offset, 42*18);
    enqueue(4*18 + x_offset, 45*18);

    enqueue(1001, 0); // up
    enqueue(16*18 + x_offset, 50*18);
    enqueue(1002, 0); // down
    enqueue(17*18 + x_offset, 51*18);
    enqueue(18*18 + x_offset, 52*18);

    enqueue(1001, 0); // up
    enqueue(17*18 + x_offset, 22*18);
    enqueue(1002, 0); // down
    enqueue(18*18 + x_offset, 18*18);
    enqueue(17*18 + x_offset, 16*18);
    enqueue(15*18 + x_offset, 15*18);
    enqueue(13*18 + x_offset, 16*18);
    enqueue(13*18 + x_offset, 17*18);
    enqueue(14*18 + x_offset, 18*18);
    enqueue(13*18 + x_offset, 22*18);

    enqueue(1001, 0); // up
    enqueue(12*18 + x_offset, 14*18);
    enqueue(1002, 0); // down
    enqueue(10*18 + x_offset, 15*18);
    enqueue(9*18 + x_offset, 15*18);
    enqueue(13*18 + x_offset, 12*18);

    enqueue(1001, 0); // up
    enqueue(14*18 + x_offset, 29*18);
    enqueue(1002, 0); // down
    enqueue(19*18 + x_offset, 29*18);
    enqueue(19*18 + x_offset, 28*18);
    enqueue(14*18 + x_offset, 28*18);

    enqueue(1001, 0); // up
    ProcessQueue();
    x_last = 0;
    y_last = 4095;
} // snoopy


// draw Bart Simpson
void bart() {
    enqueue(1001, 0); // up
    enqueue(125*5,50.2*5);
    enqueue(1002, 0); // down
    enqueue(125*5,30*5);
    enqueue(130*5,20*5);
    enqueue(120*5,10*5);
    enqueue(90*5,10*5);
    enqueue(80*5,20*5);
    enqueue(70*5,20*5);
    enqueue(70*5,25*5);
    enqueue(85*5,40*5);
    enqueue(20*5,50*5);
    enqueue(30*5,70*5);
    enqueue(40*5,80*5);

    enqueue(1001, 0); // up
    enqueue(130*5,70*5);
    enqueue(1002, 0); // down
    enqueue(180*5,130*5);
    enqueue(170*5,130*5);
    enqueue(170*5,145*5);
    enqueue(160*5,140*5);
    enqueue(160*5,150*5);
    enqueue(150*5,140*5);
    enqueue(150*5,160*5);
    enqueue(140*5,150*5);
    enqueue(140*5,170*5);
    enqueue(120*5,160*5);
    enqueue(120*5,170*5);
    enqueue(110*5,165*5);
    enqueue(110*5,180*5);
    enqueue(100*5,170*5);
    enqueue(100*5,190*5);
    enqueue(80*5,160*5);
    enqueue(70*5,140*5);
    enqueue(60*5,130*5);
  
    enqueue(1001, 0); // up
    enqueue(90*5,40*5);
    enqueue(1002, 0); // down
    enqueue(100*5,40*5);
    enqueue(80*5,30*5);
    enqueue(90*5,40*5);

    enqueue(1001, 0); // up
    enqueue(50*5,80*5);
    enqueue(1002, 0); // down
    enqueue(40*5,80*5);
    enqueue(35*5,90*5);
    enqueue(40*5,100*5);
    enqueue(50*5,105*5);
    enqueue(60*5,100*5);
    enqueue(70*5,100*5);

    enqueue(1001, 0); // up
    enqueue(50*5,105*5);
    enqueue(1002, 0); // down
    enqueue(50*5,120*5);
    enqueue(60*5,130*5);
    enqueue(70*5,130*5);
    enqueue(80*5,120*5);
    enqueue(70*5,110*5);
    enqueue(70*5,100*5);
    enqueue(75*5,90*5);
    enqueue(90*5,80*5);
    enqueue(100*5,80*5);
    enqueue(110*5,90*5);
    enqueue(110*5,110*5);
    enqueue(100*5,120*5);
    enqueue(80*5,120*5);

    enqueue(1001, 0); // up
    enqueue(125*5,70*5);
    enqueue(1002, 0); // down
    enqueue(135*5,70*5);
    enqueue(140*5,60*5);
    enqueue(130*5,50*5);
    enqueue(120*5,50*5);
    enqueue(115*5,55*5);

    enqueue(1001, 0); // up
    enqueue(125*5,65*5);
    enqueue(1002, 0); // down
    enqueue(130*5,60*5);
    enqueue(130*5,55*5);
    enqueue(125*5,55*5);
    enqueue(120*5,60*5);

    enqueue(1001, 0); // up
    enqueue(90*5,10*5);
    enqueue(1002, 0); // down
    enqueue(80*5,0*5);

    enqueue(1001, 0); // up
    enqueue(130*5,20*5);
    enqueue(1002, 0); // down
    enqueue(150*5,0*5);
 
    enqueue(1001, 0); // up
    enqueue(49*5,110*5);
    enqueue(1002, 0); // down
    enqueue(51*5,110*5);
    enqueue(1001, 0); // up
    enqueue(50*5,109*5);
    enqueue(1002, 0); // down
    enqueue(50*5,111*5);

    enqueue(1001, 0); // up
    enqueue(79*5,100*5);
    enqueue(1002, 0); // down
    enqueue(81*5,100*5);
    enqueue(1001, 0); // up
    enqueue(80*5,99*5);
    enqueue(1002, 0); // down
    enqueue(80*5,101*5);


    enqueue(1001, 0); // up
    ProcessQueue();
    x_last = 0;
    y_last = 4095;
} // bart


// Queue Setup
const int ringsize = 200;  //~415 max on the 2K SRAM UNO storage
int head = 0;
int tail = 0;
int xRing[ringsize];
int yRing[ringsize];
int xd;                       // the X coord from the queue
int yd;                       // the Y coord from the queue

// Put something into the buffer. Returns 0 when the buffer was full,
// 1 when the stuff was put sucessfully into the buffer
int enqueue (int xVal, int yVal) {
   int newtail = (tail + 1) % ringsize;
   if (newtail == head) {
	// Buffer is full, do nothing
	return 0;
   }
   else {
	xRing[tail] = xVal;
	yRing[tail] = yVal;
	tail = newtail;
	return 1;
   }
}

// Return number of elements in the queue.
int queuelevel () {
    return tail - head + (head > tail ? ringsize : 0);
}

// Get something from the queue. -1 will be returned if the queue
// is empty
int dequeueX () {
   if (head == tail) {
	return -1;
   }
   else {
	int val = xRing[head];
	head  = (head + 1) % ringsize;
	return val;
   }
} 

int dequeueY () {
   if (head == tail) {
	return -1;
   }
   else {
	int val = yRing[head];
	//head  = (head + 1) % ringsize;
	return val;
   }
} 

void dequeueXY( int &xd, int &dy) {
   if (head == tail) {
     xd = -1;
     yd = -1;
   } else {
     xd = xRing[head];
     yd = yRing[head];
     head  = (head + 1) % ringsize;
   }
} 


void ProcessQueue() {
  // JRK and position control setup
  #define MAX_DIST 4095         // the maximum value for the X or Y position (min is zero)
  #define MAX_TIME 3000         // the time it tales to move the max dist - in milliseconds
  //#define MOVE_DELAY 2000        // the time to wait between moves
  #define MOVE_DELAY 200        // the time to wait between moves

  int x1;                       // the current X Position
  int y1;                       // the current Y Position
  int x2 = x_last;                // the X position we want to move to
  int y2 = y_last;                // the Y position we want to move to
  int xPos;                     // the X position we will be moving to in this step
  int yPos;                     // the Y position we will be moving to in this step
  int xPrev;                    // the X position from the previous step
  int yPrev;                    // the Y position from the previous step

  long dX;                      // the X distance we need to move - may be negative
  long dY;                      // the Y distance we need to move - may be negative
  long dMax;                    // the larger of the X and Y distance to move - always positive (abs)
  long dT;                      // the amount of time we will need to make the move
  long elapsed_time;            // the elapsed time into the current move

  unsigned long tStart;         // the time we started the move
  unsigned long tStop;          // the time we will stop the move
  unsigned long tNow;           // the current millis() time

  boolean move_done = true;     // flag to note whether the move is complete 

  tStop = millis();

  while (queuelevel() > 0) {
    if (digitalRead(PIN_BUTTON_SELECT) == LOW) {
      head = 0;
      tail = 0;
      break;
    }  

    // get the current time
    tNow = millis();
  
    // if the time was reached, but we aren't quite done, send the last move command
    if ((tNow > tStop) && (move_done == false)) {
      XMove(x2);
      YMove(y2);
      move_done = true;    
    }

    // if we are past a move + some delay, setup a new move
    if (tNow > (tStop + MOVE_DELAY)) {
      //get next X & Y from the queue
      //x2 = random(MAX_DIST+1);
      //y2 = random(MAX_DIST+1);
      dequeueXY(xd, yd);
    if (xd == -1) {
      // do nothing - no values to use
    } else if (xd == 1001) {
      //Serial.println("UP");
      Zservo.write(servo_up);
      delay(250);                      
    } else if (xd == 1002) {
      //Serial.println("DOWN");
      Zservo.write(servo_down);
      delay(250);                      
    } else {
      x1 = x2;  // set starting location for his move to the current location
      y1 = y2;
      x2 = map(xd, 0, 1000, 0, MAX_DIST);  // set the new target location
      y2 = map(yd, 0, 1000, 0, MAX_DIST);
      dX = x2 - x1;
      dY = y2 - y1;
      dMax = max(abs(dX), abs(dY));
      dT = MAX_TIME * dMax / MAX_DIST;
      tStart = millis();
      tStop = tStart + dT;
      move_done = false;
    }
  }
 
  // if we are in a move, send commands to the controller to keep it moving
  if (tNow < tStop) {
    elapsed_time = tNow - tStart;

    xPos = x1 + int(dX * elapsed_time / dT);
    // only move if we've actually changed position
    if (xPos != xPrev) {
      XMove(xPos);
      xPrev = xPos;
    }

    yPos = y1 + int(dY * elapsed_time / dT);
    // only move if we've actually changed position
    if (yPos != yPrev) {
      YMove(yPos);
      yPrev = yPos;
    }

  }


  } // while loop
} // ProcesssQueue

void setup() {
  Serial.begin(9600);

  // Joystick Shield setup
  // Specify each pin connected to a pushbutton as an input.
  // Also enable the Arduino's internal "pull-up" resistors
  // for each pushbutton we want to read--this means the shield
  // doesn't need to have resistors on it.
  // Note that when a pull-up resistor is used on a pin the
  // meaning of the values read are reversed compared to their
  // usual meanings:
  //    * HIGH = the button is not pressed
  //    * LOW = the button is pressed  
  pinMode(PIN_BUTTON_RIGHT, INPUT);  
  digitalWrite(PIN_BUTTON_RIGHT, HIGH);
  
  pinMode(PIN_BUTTON_LEFT, INPUT);  
  digitalWrite(PIN_BUTTON_LEFT, HIGH);
  
  pinMode(PIN_BUTTON_UP, INPUT);  
  digitalWrite(PIN_BUTTON_UP, HIGH);
  
  pinMode(PIN_BUTTON_DOWN, INPUT);  
  digitalWrite(PIN_BUTTON_DOWN, HIGH);
  
  pinMode(PIN_BUTTON_SELECT, INPUT);  
  digitalWrite(PIN_BUTTON_SELECT, HIGH);  
  
  x_last = 2048;  // start with the X & Y Axis in the center (they go from 0-4095)
  y_last = 2048;
  XMove(x_last);
  YMove(y_last);
  
  Zservo.attach(9);         // attaches the servo on pin 9 to the servo object 
  Zservo.write(servo_up);   // move the pen up

}


void loop() {
  
  if (0) {
  // Print the current values of the inputs (joystick and
  // buttons) to the console.
  Serial.print("l:");
  Serial.print(digitalRead(PIN_BUTTON_LEFT));
  Serial.print(" ");
  
  Serial.print("r:");
  Serial.print(digitalRead(PIN_BUTTON_RIGHT));
  Serial.print(" ");
  
  Serial.print("u:");
  Serial.print(digitalRead(PIN_BUTTON_UP));
  Serial.print(" ");
  
  Serial.print("d:");
  Serial.print(digitalRead(PIN_BUTTON_DOWN));
  Serial.print(" ");

  Serial.print("x:");
  Serial.print(analogRead(PIN_ANALOG_X));
  Serial.print(" ");
  
  Serial.print("y:");
  Serial.print(analogRead(PIN_ANALOG_Y));
  Serial.print(" ");  
  
  Serial.print("s:");
  Serial.print(digitalRead(PIN_BUTTON_SELECT));
  Serial.print(" ");
  
  Serial.println();
  }
  
  
  if ((digitalRead(PIN_BUTTON_RIGHT) == LOW) && (digitalRead(PIN_BUTTON_UP) == LOW)) {  // if the right button is pressed, do a wipeout
    wipeout();
    delay(500);
  } else if ((digitalRead(PIN_BUTTON_RIGHT) == LOW) && (digitalRead(PIN_BUTTON_LEFT) == LOW)) {
    snoopy();
    delay(500);
  } else if ((digitalRead(PIN_BUTTON_RIGHT) == LOW) && (digitalRead(PIN_BUTTON_DOWN) == LOW)) {
    bart();
    delay(500);
  } else if (digitalRead(PIN_BUTTON_LEFT) == LOW) {  // The Left button centers the arm
    x_last = 2048;
    y_last = 2048;
    XMove(x_last);
    YMove(y_last);
  } else if (digitalRead(PIN_BUTTON_UP) == LOW) {  // if the up button is pressed, lift the servo
    Zservo.write(servo_up);
    delay(250);                      
  } else if (digitalRead(PIN_BUTTON_DOWN) == LOW) {  // if the down button is pressed, lower the servo
    Zservo.write(servo_down);
    delay(250);                      
  } else { // do the normal joystick move
    // X Axis move
    x_joy = map(analogRead(PIN_ANALOG_X), 0, 1023, -40, 40);
    if ((x_joy > 2) || (x_joy < -2)) {  // adding a dead band in the middle since they joysticks are not always exactly zero at the center
      x_last = min(4095, max(0, x_last + x_joy));
      XMove(x_last);
    }

    //Y Axis move
    y_joy = map(analogRead(PIN_ANALOG_Y), 0, 1023, -40, 40);
    if ((y_joy > 2) || (y_joy < -2)) {  // adding a dead band in the middle since they joysticks are not always exactly zero at the center
      y_last = min(4095, max(0, y_last + y_joy));
      YMove(y_last);
    }

    delay(30); // give the commands a chance to get transmitted at 9600 baud
  } // else
   
} //loop

